@@ -1,18 +1,138 @@
 from OpenFlows.Domain.ModelingElements.NetworkElements import IActiveElementsInput, IBaseLinksInput, IBasePolygonInput, IBasePolygonsInput, INetworkElements, IPointNodesInput
 from OpenFlows.Water.Domain import IWaterModel
-from OpenFlows.Water.Domain.ModelingElements.NetworkElements import IBaseDirectedNodesInput, IBaseNodesInput, IBasePumpsInput, IBaseValvesInput, ICheckValveElementsInput, IConventionalTanksInput, ICustomerMetersInput, IDemandNodesInput, IFireFlowNodesInput, IFlowControlValvesInput, IGeneralPurposeValves, IGeneralPurposeValvesInput, IHydrantsInput, IHydroTanksInput, IIsolationValveElementsInput, IJunctionsInput, ILateralsInput, IPhysicalNodeElementsInput, IPipes, IPressureBreakingValves, IPressureBreakingValvesInput, IPressureSustainingValvesInput, IPressureValvesInput, IPumpStations, IPumpStationsInput, IPumps, IPumpsInput, IReservoirs, ISCADAElements, ISCADAElementsInput, ITanks, ITanksInput, ITaps, IThrottleControlValvesInput, IVSPBsInput, IWaterQualityNodesInput, IWaterZoneableNetworkElementsInput
+from OpenFlows.Water.Domain.ModelingElements.NetworkElements import IBaseDirectedNodesInput, IBaseNodesInput, IBasePumpsInput, IBaseValvesInput, ICheckValveElementsInput, IConventionalTanksInput, ICustomerMetersInput, IDemandNodesInput, IFireFlowNodesInput, IFlowControlValvesInput, IGeneralPurposeValves, IGeneralPurposeValvesInput, IHydrantsInput, IHydroTanksInput, IIsolationValveElementsInput, IJunctionsInput, ILateralsInput, IPhysicalNodeElementsInput, IPipes, IPressureBreakingValves, IPressureBreakingValvesInput, IPressureSustainingValvesInput, IPressureValvesInput, IPumpStations, IPumpStationsInput, IPumps, IPumpsInput, IReservoirs, ISCADAElements, ISCADAElementsInput, ITanks, ITanksInput, ITaps, IThrottleControlValvesInput, IVSPBsInput, IWaterQualityElementsInput, IWaterQualityNodesInput, IWaterZoneableNetworkElementsInput
 import numpy as np
 import pandas as pd
+import networkx as nx
 from typing import Any, List, Type
 
 
 class NetworkInput:
+    # region Fields
     __waterModel: IWaterModel
+    # endregion
 
+    # region Constructor
     def __init__(self, water_model: IWaterModel) -> None:
         self.__waterModel = water_model
         pass
 
+    def __repr__(self) -> str:
+        line1 = f"{__class__.__name__}: {self.__waterModel}."
+        line2 = f"Pipe Count: {self.__waterModel.Network.Pipes.Count}"
+        line3 = f"Junction Count: {self.__waterModel.Network.Junctions.Count}"
+        return f"{line1} {line2} {line3}."
+    # endregion
+
+    # region Public Methods
+    def get_networkx_graph(self, laterals: bool = False) -> nx.Graph:
+        columns = ["Id", "Label"] if laterals else [
+            "Id", "Label", "Diameter", "IsActive"]
+
+        links_df = self.pipe_df[columns].copy()
+        if laterals:
+            links_df.append(self.lateral_df[columns])
+
+        graph: nx.Graph = nx.from_pandas_edgelist(
+            df=links_df,
+            source="StartNodeId",
+            target="StopNodeId",
+            edge_attr=columns)
+
+        return graph
+
+    # endregion // Public Methods
+
+    # region Public Properties (Network Elements DF)
+
+    @property
+    def pipe_df(self) -> pd.DataFrame:
+        return self.__get_pipe_input(self.__waterModel.Network.Pipes)
+
+    @property
+    def lateral_df(self) -> pd.DataFrame:
+        return self.__get_lateral_input(self.__waterModel.Network.Laterals)
+
+    @property
+    def junction_df(self) -> pd.DataFrame:
+        return self.__get_junction_input(self.__waterModel.Network.Junctions)
+
+    @property
+    def hydrant_df(self) -> pd.DataFrame:
+        return self.__get_hydrant_input(self.__waterModel.Network.Hydrants)
+
+    @property
+    def tank_df(self) -> pd.DataFrame:
+        return self.__get_tank_input(self.__waterModel.Network.Tanks)
+
+    @property
+    def reservoir_df(self) -> pd.DataFrame:
+        return self.__get_reservoir_input(self.__waterModel.Network.Reservoirs)
+
+    @property
+    def tap_df(self) -> pd.DataFrame:
+        return self.__get_tap_input(self.__waterModel.Network.Taps)
+
+    @property
+    def pump_df(self) -> pd.DataFrame:
+        return self.__get_pump_input(self.__waterModel.Network.Pumps)
+
+    @property
+    def pump_stn_df(self) -> pd.DataFrame:
+        return self.__get_pump_stn_input(self.__waterModel.Network.PumpStations)
+
+    @property
+    def customer_meter_df(self) -> pd.DataFrame:
+        return self.__get_customer_meter_input(self.__waterModel.Network.CustomerMeters)
+
+    @property
+    def scada_elem_df(self) -> pd.DataFrame:
+        return self.__get_scada_elem_input(self.__waterModel.Network.SCADAElements)
+
+    @property
+    def vspb_df(self) -> pd.DataFrame:
+        return self.__get_vspb_input(self.__waterModel.Network.VSPBs)
+
+    @property
+    def prv_df(self) -> pd.DataFrame:
+        return self.__get_prv_input(self.__waterModel.Network.PRVs)
+
+    @property
+    def psv_df(self) -> pd.DataFrame:
+        return self.__get_psv_input(self.__waterModel.Network.PSVs)
+
+    @property
+    def pbv_df(self) -> pd.DataFrame:
+        return self.__get_pbv_input(self.__waterModel.Network.PBVs)
+
+    @property
+    def fcv_df(self) -> pd.DataFrame:
+        return self.__get_fcv_input(self.__waterModel.Network.FCVs)
+
+    @property
+    def tcv_df(self) -> pd.DataFrame:
+        return self.__get_tcv_input(self.__waterModel.Network.TCVs)
+
+    @property
+    def gpv_df(self) -> pd.DataFrame:
+        return self.__get_gpv_input(self.__waterModel.Network.GPVs)
+
+    @property
+    def iso_valve_df(self) -> pd.DataFrame:
+        return self.__get_iso_valve_input(self.__waterModel.Network.IsolationValves)
+
+    @property
+    def hydro_tank_df(self) -> pd.DataFrame:
+        return self.__get_hydro_tank_input(self.__waterModel.Network.HydropneumaticTanks)
+
+    @property
+    def check_valve_df(self) -> pd.DataFrame:
+        return self.__get_check_valve_input(self.__waterModel.Network.CheckValves)
+
+    # endregion // Public Properties
+
+    # region Private methods
+
     def __dict_to_value(self, series: pd.Series, data_type: Type) -> pd.Series:
         series = series.apply(lambda d: d.Value)
         if data_type:
@@ -24,10 +144,9 @@ class NetworkInput:
 
     def __get_elements_input(self, elements: INetworkElements) -> pd.DataFrame:
         df = pd.DataFrame()
-        # TODO
-        # df["Label"] = elements.Input.Labels()
-        # df["Id"] = df["Label"].apply(lambda d: d.Key).astype(np.int64)
-        # df["Label"] = df["Label"].apply(lambda d: d.Value).astype("string")
+        df["Label"] = elements.Labels()
+        df["Id"] = df["Label"].apply(lambda d: d.Key).astype(pd.Int64Dtype())
+        df["Label"] = df["Label"].apply(lambda d: d.Value).astype("string")
         return df
 
     def __get_physical_elevation_input(self, elements: IPhysicalNodeElementsInput, df: pd.DataFrame) -> pd.DataFrame:
@@ -63,15 +182,15 @@ class NetworkInput:
 
         return df
 
-    def __get_polygons_geometry(self, elements: IBasePolygonInput, df: pd.DataFrame) -> pd.DataFrame:
-        df["Geometry"] = elements.GetRings()
-        df["Geometry"] = self._dict_to_value(df["Geometry"], None)
+    def __get_polygons_geometry(self, elements: IBasePolygonsInput, df: pd.DataFrame) -> pd.DataFrame:
+        df["Geometry"] = elements.Geometries()
+        df["Geometry"] = self.__dict_to_value(df["Geometry"], None)
 
         df["Geometry"] = df["Geometry"].apply(
             lambda pts: [[p.X, p.Y] for p in pts]).tolist()
         return df
 
-    def __get_water_quality_node_input(self, elements: IWaterQualityNodesInput, df: pd.DataFrame) -> pd.DataFrame:
+    def __get_water_quality_node_input(self, elements: IWaterQualityElementsInput, df: pd.DataFrame) -> pd.DataFrame:
         df["InitAge"] = elements.InitialAge()
         df["InitAge"] = self.__dict_to_value(df["InitAge"], float)
 
@@ -84,7 +203,8 @@ class NetworkInput:
 
     def __get_installation_year_input(self, elements: Any, df: pd.DataFrame) -> pd.DataFrame:
         df["InstallYr"] = elements.InstallationYears()
-        df["InstallYr"] = self.__dict_to_value(df["InstallYr"], np.int64)
+        df["InstallYr"] = self.__dict_to_value(
+            df["InstallYr"], pd.Int64Dtype())
         return df
 
     def __get_minor_loss_node_input(self, elements: Any, df: pd.DataFrame) -> pd.DataFrame:
@@ -107,7 +227,8 @@ class NetworkInput:
 
     def __get_hammer_valve_type_input(self, elements: Any, df: pd.DataFrame) -> pd.DataFrame:
         df["ValveType"] = elements.ValveTypes()
-        df["ValveType"] = self.__dict_to_value(df["ValveType"], np.int64)
+        df["ValveType"] = self.__dict_to_value(
+            df["ValveType"], pd.Int64Dtype())
         return df
 
     def __get_demand_node_input(self, elements: IDemandNodesInput) -> pd.DataFrame:
@@ -139,12 +260,12 @@ class NetworkInput:
         df["StartNode"] = elements.StartNodes()
         df["StartNode"] = self.__dict_to_value(df["StartNode"], None)
         df["StartNodeId"] = df["StartNode"].apply(
-            lambda n: n.Id).astype(np.int64)
+            lambda n: n.Id).astype(pd.Int64Dtype())
 
         df["StopNode"] = elements.StopNodes()
         df["StopNode"] = self.__dict_to_value(df["StopNode"], None)
         df["StopNodeId"] = df["StopNode"].apply(
-            lambda n: n.Id).astype(np.int64)
+            lambda n: n.Id).astype(pd.Int64Dtype())
 
         df["IsUDLength"] = elements.IsUserDefinedLengths()
         df["IsUDLength"] = self.__dict_to_value(df["IsUDLength"], bool)
@@ -171,7 +292,7 @@ class NetworkInput:
         df["AssocElem"] = elements.AssociatedElements()
         df["AssocElem"] = self.__dict_to_value(df["AssocElem"], None)
         df["AssocElemId"] = df["AssocElem"].apply(
-            lambda n: n.Id).astype(np.int64)
+            lambda n: n.Id if n else None).astype(pd.Int64Dtype())
         return df
 
     # region Base Elements Input
@@ -417,7 +538,8 @@ class NetworkInput:
             df["InitSpeedFactor"], float)
 
         df["InitStatus"] = elements.InitialStatus()
-        df["InitStatus"] = self.__dict_to_value(df["InitStatus"], np.int64)
+        df["InitStatus"] = self.__dict_to_value(
+            df["InitStatus"], pd.Int64Dtype())
 
         # TODO: double check the fields
         return df
@@ -433,13 +555,13 @@ class NetworkInput:
         df["PumpDefinition"] = self.__dict_to_value(
             df["PumpDefinition"], None)
         df["PumpDefinitionId"] = df["PumpDefinition"].apply(
-            lambda p: p.Id).astype(np.int64)
+            lambda p: p.Id if p else None).astype(pd.Int64Dtype())
 
         df["ControlNode"] = elements.ControlNodes()
         df["ControlNode"] = self.__dict_to_value(
             df["ControlNode"], None)
         df["ControlNodeId"] = df["ControlNode"].apply(
-            lambda p: p.Id).astype(np.int64)
+            lambda p: p.Id if p else None).astype(pd.Int64Dtype())
 
         df["TgtHGL"] = elements.TargetHydraulicGrades()
         df["TgtHGL"] = self.__dict_to_value(
@@ -451,13 +573,13 @@ class NetworkInput:
 
         df["NumLagPumps"] = elements.NumberOfLagPumps()
         df["NumLagPumps"] = self.__dict_to_value(
-            df["NumLagPumps"], np.int64)
+            df["NumLagPumps"], pd.Int64Dtype())
 
         df["CtrlNodeSucSide"] = elements.ControlNodeOnSuctionSide()
         df["CtrlNodeSucSide"] = self.__dict_to_value(
             df["CtrlNodeSucSide"], None)
         df["CtrlNodeSucSideId"] = df["CtrlNodeSucSide"].apply(
-            lambda p: p.Id).astype(np.int64)
+            lambda p: p.Id if p else None).astype(pd.Int64Dtype())
 
         df["TgtFlow"] = elements.TargetFlows()
         df["TgtFlow"] = self.__dict_to_value(
@@ -482,13 +604,15 @@ class NetworkInput:
     def __get_customer_meter_input(self, elements: ICustomerMetersInput) -> pd.DataFrame:
         df = self.__get_elements_input(elements)
         df = self.__get_point_node_input(elements, df)
+        df = self.__get_physical_elevation_input(elements, df)
 
         df["Demand"] = elements.BaseDemands()
         df["Demand"] = self.__dict_to_value(df["Demand"], float)
 
         df["Pattern"] = elements.DemandPatterns()
         df["Pattern"] = self.__dict_to_value(df["Pattern"], None)
-        df["PatternId"] = df["Pattern"].apply(lambda p: p.Id).astype(np.int64)
+        df["PatternId"] = df["Pattern"].apply(
+            lambda p: p.Id if p else None).astype(pd.Int64Dtype())
 
         df["StartDemandDist"] = elements.StartDemandDistributions()
         df["StartDemandDist"] = self.__dict_to_value(
@@ -497,7 +621,7 @@ class NetworkInput:
         df["AssocElem"] = elements.AssociatedElements()
         df["AssocElem"] = self.__dict_to_value(df["AssocElem"], None)
         df["AssocElemId"] = df["AssocElem"].apply(
-            lambda c: c.Id).astype(np.int64)
+            lambda c: c.Id if c else None).astype(pd.Int64Dtype())
 
         df["UnitDemand"] = elements.UnitDemands()
         df["UnitDemand"] = self.__dict_to_value(df["UnitDemand"], float)
@@ -505,7 +629,7 @@ class NetworkInput:
         df["UnitDmdPattern"] = elements.UnitDemandPatterns()
         df["UnitDmdPattern"] = self.__dict_to_value(df["UnitDmdPattern"], None)
         df["UnitDmdPatternId"] = df["UnitDmdPattern"].apply(
-            lambda p: p.Id).astype(np.int64)
+            lambda p: p.Id if p else None).astype(pd.Int64Dtype())
 
         df["NumUnitDmd"] = elements.UnitDemands()
         df["NumUnitDmd"] = self.__dict_to_value(df["NumUnitDmd"], float)
@@ -521,12 +645,13 @@ class NetworkInput:
 
         df["TgtElem"] = elements.TargetElements()
         df["TgtElem"] = self.__dict_to_value(df["TgtElem"], None)
-        df["TgtElemId"] = df["TgtElem"].apply(lambda e: e.Id).astype(np.int64)
+        df["TgtElemId"] = df["TgtElem"].apply(
+            lambda e: e.Id if e else None).astype(pd.Int64Dtype())
 
         df["HistSignal"] = elements.HistoricalSignals()
         df["HistSignal"] = self.__dict_to_value(df["HistSignal"], None)
         df["HistSignalId"] = df["HistSignal"].apply(
-            lambda s: s.Id).astype(np.int64)
+            lambda s: s.Id if s else None).astype(pd.Int64Dtype())
         return df
     # endregion
 
@@ -594,7 +719,8 @@ class NetworkInput:
 
         df["RefPipe"] = elements.ReferencedPipes()
         df["RefPipe"] = self.__dict_to_value(df["RefPipe"], None)
-        df["RefPipeId"] = df["RefPipe"].apply(lambda p: p.Id).astype(np.int64)
+        df["RefPipeId"] = df["RefPipe"].apply(
+            lambda p: p.Id if p else None).astype(pd.Int64Dtype())
 
         df["Diameter"] = elements.ValveDiameters()
         df["Diameter"] = self.__dict_to_value(df["Diameter"], float)
@@ -635,89 +761,4 @@ class NetworkInput:
 
     # endregion
 
-# region Network Elements DF
-
-    @property
-    def pipe_df(self) -> pd.DataFrame:
-        return self.__get_pipe_input(self.__waterModel.Network.Pipes)
-
-    @property
-    def lateral_df(self) -> pd.DataFrame:
-        return self.__get_lateral_input(self.__waterModel.Network.Laterals)
-
-    @property
-    def junction_df(self) -> pd.DataFrame:
-        return self.__get_junction_input(self.__waterModel.Network.Junctions)
-
-    @property
-    def hydrant_df(self) -> pd.DataFrame:
-        return self.__get_hydrant_input(self.__waterModel.Network.Hydrants)
-
-    @property
-    def tank_df(self) -> pd.DataFrame:
-        return self.__get_tank_input(self.__waterModel.Network.Tanks)
-
-    @property
-    def reservoir_df(self) -> pd.DataFrame:
-        return self.__get_reservoir_input(self.__waterModel.Network.Reservoirs)
-
-    @property
-    def tap_df(self) -> pd.DataFrame:
-        return self.__get_tap_input(self.__waterModel.Network.Taps)
-
-    @property
-    def pump_df(self) -> pd.DataFrame:
-        return self.__get_pump_input(self.__waterModel.Network.Pumps)
-
-    @property
-    def pump_stn_df(self) -> pd.DataFrame:
-        return self.__get_pump_stn_input(self.__waterModel.Network.PumpStations)
-
-    @property
-    def customer_meter_df(self) -> pd.DataFrame:
-        return self.__get_customer_meter_input(self.__waterModel.Network.CustomerMeters)
-
-    @property
-    def scada_elem_df(self) -> pd.DataFrame:
-        return self.__get_scada_elem_input(self.__waterModel.Network.SCADAElements)
-
-    @property
-    def vspb_df(self) -> pd.DataFrame:
-        return self.__get_vspb_input(self.__waterModel.Network.VSPBs)
-
-    @property
-    def prv_df(self) -> pd.DataFrame:
-        return self.__get_prv_input(self.__waterModel.Network.PRVs)
-
-    @property
-    def psv_df(self) -> pd.DataFrame:
-        return self.__get_psv_input(self.__waterModel.Network.PSVs)
-
-    @property
-    def pbv_df(self) -> pd.DataFrame:
-        return self.__get_pbv_input(self.__waterModel.Network.PBVs)
-
-    @property
-    def fcv_df(self) -> pd.DataFrame:
-        return self.__get_fcv_input(self.__waterModel.Network.FCVs)
-
-    @property
-    def tcv_df(self) -> pd.DataFrame:
-        return self.__get_tcv_input(self.__waterModel.Network.TCVs)
-
-    @property
-    def gpv_df(self) -> pd.DataFrame:
-        return self.__get_gpv_input(self.__waterModel.Network.GPVs)
-
-    @property
-    def iso_valve_df(self) -> pd.DataFrame:
-        return self.__get_iso_valve_input(self.__waterModel.Network.IsolationValves)
-
-    @property
-    def hydro_tank_df(self) -> pd.DataFrame:
-        return self.__get_hydro_tank_input(self.__waterModel.Network.HydropneumaticTanks)
-
-    @property
-    def check_valve_df(self) -> pd.DataFrame:
-        return self.__get_check_valve_input(self.__waterModel.Network.CheckValves)
-# endregion
+    # endregion // private methods
